from math import floor
import random
from typing import List, TYPE_CHECKING, Tuple, cast

import torch
from transformers.modeling_outputs import BaseModelOutputWithPastAndCrossAttentions
from transformers.models.t5.modeling_t5 import T5Stack

from mlit.models.common import InputDescription, OnnxModelConverterHelper

if TYPE_CHECKING:
    from onnxruntime import InferenceSession
    import torch.nn


class T5DecoderNoHistoryDescription(OnnxModelConverterHelper['T5Stack']):
    '''
    https://github.com/huggingface/transformers/blob/master/src/transformers/models/t5/modeling_t5.py#L836-L850
    '''
    subname = 'decoder_no_history'

    inputs = {
        'input_ids': InputDescription(
            dynamic_axis={0: 'batch_size', 1: 'decoder_seq_len'}
        ),
        'attention_mask': InputDescription(
            dynamic_axis={0: 'batch_size', 1: 'decoder_seq_len'}
        ),
        'encoder_hidden_states': InputDescription(
            dynamic_axis={0: 'batch_size', 1: 'encoder_seq_len'}
        ),
        'encoder_attention_mask': InputDescription(
            dynamic_axis={0: 'batch_size', 1: 'encoder_seq_len'}
        )
    }
    outputs = {
        'logits': InputDescription(
            dynamic_axis={0: 'batch_size', 1: 'output_seq_len'}
        )
    }
    default_forward_args = {
        'inputs_embeds': None,
        'head_mask': None,
        # 'cross_attn_layer_head_mask': None,
        'past_key_values': None,
        'use_cache': False,
        'output_attentions': False,
        'output_hidden_states': False,
        'return_dict': True
    }

    _lm_head: torch.nn.Linear

    def __init__(self, encoder: 'T5Stack', lm_head: torch.nn.Linear):
        super().__init__(encoder)
        self._lm_head = lm_head

    def forward(self, *args, **kwargs):
        result = super().forward(*args, **kwargs)

        if self.model.config.tie_word_embeddings:
            d_model = self.model.config.d_model  # type: ignore
            # Rescale output before projecting on vocab
            # See https://github.com/tensorflow/mesh/blob/fa19d69eafc9a482aff0b59ddd96b025c0cb207d/mesh_tensorflow/transformer/transformer.py#L586
            state = result.last_hidden_state * \
                (d_model ** -0.5)
        else:
            state = result.last_hidden_state

        return [self._lm_head(state)]

    def sample_inputs(
        self,
        batch_size: int = 2,
        encoder_sequence_length: int = 50,
        decoder_sequence_length: int = 20
    ) -> Tuple:

        vocab_size: int = self.model.config.vocab_size  # type: ignore
        hidden_size: int = self.model.config.d_model  # type: ignore

        input_ids = torch.randint(low=0,
                                  high=vocab_size - 1,
                                  size=(batch_size, decoder_sequence_length),
                                  dtype=torch.int64)

        attention_mask = torch.ones(
            [batch_size, decoder_sequence_length],
            dtype=torch.int64)

        if decoder_sequence_length >= 2:
            # mask one word in a random position
            padding_position = random.randint(1, decoder_sequence_length - 1)
            attention_mask[:, padding_position:] = 0

        encoder_hidden_states = torch.rand(batch_size,
                                           encoder_sequence_length,
                                           hidden_size,
                                           dtype=torch.float32)

        encoder_attention_mask = torch.ones(
            [batch_size, encoder_sequence_length],
            dtype=torch.int64)

        return (
            input_ids,
            attention_mask,
            encoder_hidden_states,
            encoder_attention_mask
        )


class T5DecoderNoHistoryInferenceSessionWrapper(T5Stack):
    def __init__(self, onnx_model: 'InferenceSession', config):
        super(T5Stack, self).__init__(config)
        self._onnx_model = onnx_model

    def forward(self,
                input_ids=None,
                attention_mask=None,
                encoder_hidden_states=None,
                encoder_attention_mask=None,
                return_dict=None,
                **kwargs):

        if attention_mask is None:
            attention_mask = torch.ones(
                input_ids.shape,
                dtype=torch.int64)
        [logits] = self._onnx_model.run(
            None,
            {
                'input_ids': input_ids.cpu().numpy(),
                'attention_mask': attention_mask.cpu().numpy(),
                'encoder_hidden_states': encoder_hidden_states.cpu().numpy(),
                'encoder_attention_mask': encoder_attention_mask.cpu().numpy()
            }
        )
        logits = torch.FloatTensor(logits)

        if not return_dict:
            return tuple(logits)
        return BaseModelOutputWithPastAndCrossAttentions(
            last_hidden_state=logits,
        )


class T5DecoderFirstStepHistoryDescription(T5DecoderNoHistoryDescription):
    subname = 'decoder_history_first_step'

    def __init__(self, decoder: T5Stack, lm_head: torch.nn.Linear):
        super().__init__(decoder, lm_head)

        num_decoder_layers: int = decoder.config.num_decoder_layers  # type: ignore

        # self.inputs['input_ids'].dynamic_axis = {0: 'batch_size'}

        extra_outputs = {}
        for i in range(num_decoder_layers):
            for j in range(4):
                extra_outputs[f'past_key_value_{i}_{j}'] = InputDescription(
                    dynamic_axis={0: 'batch_size', 1: 'decoder_seq_len'}
                )

        self.outputs = {
            **self.outputs,
            **extra_outputs
        }

        self.default_forward_args['use_cache'] = True

    def forward(self, *args, **kwargs):
        result = OnnxModelConverterHelper.forward(self, *args, **kwargs)

        if self.model.config.tie_word_embeddings:
            d_model = self.model.config.d_model  # type: ignore
            # Rescale output before projecting on vocab
            # See https://github.com/tensorflow/mesh/blob/fa19d69eafc9a482aff0b59ddd96b025c0cb207d/mesh_tensorflow/transformer/transformer.py#L586
            state = result.last_hidden_state * (d_model ** -0.5)
        else:
            state = result.last_hidden_state

        # unravel past key values
        past_key_values = result.past_key_values
        past_key_values_outputs = []
        for i, item in enumerate(past_key_values):
            for j, inner_item in enumerate(item):
                past_key_values_outputs += [inner_item]

        return [self._lm_head(state)] + past_key_values_outputs

    def sample_inputs(
        self,
        batch_size: int = 2,
        encoder_sequence_length: int = 50,
        decoder_sequence_length: int = 20
    ) -> Tuple:
        inputs = T5DecoderNoHistoryDescription.sample_inputs(
            self, batch_size, encoder_sequence_length, decoder_sequence_length)

        a, b, c, d = inputs
        # when past key values are present, only the last input is interesting
        inputs = (a[:, -1:], b[:, -1:], c, d)

        return inputs


def from_unravelled_past_key_values(
    unraveled_past_key_values: List[torch.FloatTensor]
) -> Tuple[Tuple[torch.FloatTensor]]:
    past_key_values = {}

    for idx, kv in enumerate(unraveled_past_key_values):
        i = floor(idx / 4)
        j = idx % 4

        if i not in past_key_values:
            past_key_values[i] = {}
        past_key_values[i][j] = kv

    return tuple([
        tuple([
            past_key_values[i][j]
            for j in sorted(past_key_values[i].keys())
        ])
        for i in sorted(past_key_values.keys())
    ])


def to_unravelled_past_key_values(past_key_values):
    # unravel past key values
    past_key_values_outputs = []
    for i, item in enumerate(past_key_values):
        for j, inner_item in enumerate(item):
            past_key_values_outputs += [inner_item]
    return past_key_values_outputs


def generate_past_key_values_fields(num_layers):
    names = []
    for i in range(int(num_layers)):
        for j in range(4):
            if j < 2:
                names += [f'input_past_key_value_{i}_{j}']
            else:
                names += [f'past_key_value_{i}_{j}']
    return names


class T5DecoderHistoryDescription(T5DecoderNoHistoryDescription):
    subname = 'decoder_history'

    def __init__(self, decoder: T5Stack, lm_head: torch.nn.Linear):
        super().__init__(decoder, lm_head)

        num_decoder_layers: int = decoder.config.num_decoder_layers  # type: ignore

        # self.inputs['input_ids'].dynamic_axis = {0: 'batch_size'}

        extra_outputs = {}
        extra_inputs = {}
        for i in range(num_decoder_layers):
            for j in range(4):
                extra_outputs[f'past_key_value_{i}_{j}'] = InputDescription(
                    dynamic_axis={0: 'batch_size',
                                  2: 'decoder_seq_len' if j < 2 else 'moo'}
                )
                input_key = f'input_past_key_value_{i}_{j}' if j < 2 else f'past_key_value_{i}_{j}'
                extra_inputs[input_key] = InputDescription(
                    dynamic_axis={0: 'batch_size',
                                  2: 'decoder_seq_len' if j < 2 else 'moo'}
                )

        self.outputs = {
            **self.outputs,
            **extra_outputs
        }

        self.inputs = {
            **self.inputs,
            **extra_inputs
        }

        self.default_forward_args['use_cache'] = True
        if 'past_key_values' in self.default_forward_args:
            del self.default_forward_args['past_key_values']

    def forward(self, *args, **kwargs):

        num_decoder_layers: int = self.model.config.num_decoder_layers  # type: ignore
        offset = len(args) - (num_decoder_layers * 4)
        unraveled_past_key_values = args[offset:]

        past_key_values = from_unravelled_past_key_values(
            cast(List[torch.FloatTensor], unraveled_past_key_values))

        args = args[:offset]
        kwargs['past_key_values'] = past_key_values

        result = OnnxModelConverterHelper.forward(self, *args, **kwargs)

        if self.model.config.tie_word_embeddings:
            d_model = self.model.config.d_model  # type: ignore
            # Rescale output before projecting on vocab
            # See https://github.com/tensorflow/mesh/blob/fa19d69eafc9a482aff0b59ddd96b025c0cb207d/mesh_tensorflow/transformer/transformer.py#L586
            state = result.last_hidden_state * (d_model ** -0.5)
        else:
            state = result.last_hidden_state

        return [self._lm_head(state)] + to_unravelled_past_key_values(result.past_key_values)

    def sample_inputs(
        self,
        batch_size: int = 2,
        encoder_sequence_length: int = 50,
        decoder_sequence_length: int = 20
    ) -> Tuple:
        inputs = T5DecoderNoHistoryDescription.sample_inputs(
            self, batch_size, encoder_sequence_length, decoder_sequence_length)

        num_decoder_layers: int = self.model.config.num_decoder_layers  # type: ignore
        num_heads: int = self.model.config.num_heads  # type: ignore
        d_kv: int = self.model.config.d_kv  # type: ignore

        past_key_values = []

        for i in range(num_decoder_layers):
            for j in range(4):
                if j < 2:
                    third_dim = decoder_sequence_length
                else:
                    third_dim = encoder_sequence_length

                past_key_values += [
                    torch.rand(batch_size,
                               num_heads,
                               third_dim,
                               d_kv,
                               dtype=torch.float32)
                ]
        a, b, c, d = inputs
        # when past key values are present, only the last input is interesting
        inputs = (a[:, -1:], b[:, -1:], c, d)

        return inputs + tuple(past_key_values)


class T5DecoderInferenceSessionWrapper(T5Stack):
    def __init__(self, onnx_model_first_step: 'InferenceSession', onnx_model: 'InferenceSession', config):
        super(T5Stack, self).__init__(config)
        self._onnx_model = onnx_model
        self._onnx_model_first_step = onnx_model_first_step

    def forward(self,
                input_ids=None,
                attention_mask=None,
                encoder_hidden_states=None,
                encoder_attention_mask=None,
                past_key_values=None,
                return_dict=None,
                **kwargs):

        if attention_mask is None:
            attention_mask = torch.ones(
                input_ids.shape,
                dtype=torch.int64)

        if past_key_values is not None:
            unravelled_past_key_values = to_unravelled_past_key_values(
                past_key_values)
            unravelled_past_key_value_names = generate_past_key_values_fields(
                len(unravelled_past_key_values) / 4)
            past_key_value_inputs = {
                k: v.cpu().numpy()
                for k, v in zip(unravelled_past_key_value_names, unravelled_past_key_values)
            }

            usual_inputs = {
                'input_ids': input_ids.cpu().numpy(),
                'attention_mask': attention_mask.cpu().numpy(),
                'encoder_hidden_states': encoder_hidden_states.cpu().numpy(),
                'encoder_attention_mask': encoder_attention_mask.cpu().numpy(),
            }
            results = self._onnx_model.run(
                None,
                {
                    **usual_inputs,
                    **past_key_value_inputs
                }
            )
        else:
            results = self._onnx_model_first_step.run(
                None,
                {
                    'input_ids': input_ids.cpu().numpy(),
                    'attention_mask': attention_mask.cpu().numpy(),
                    'encoder_hidden_states': encoder_hidden_states.cpu().numpy(),
                    'encoder_attention_mask': encoder_attention_mask.cpu().numpy()
                }
            )
        results = [torch.FloatTensor(i) for i in results]
        logits = results[0]
        past_key_values = results[1:]
        past_key_values = from_unravelled_past_key_values(past_key_values)

        if not return_dict:
            return tuple(logits)
        return BaseModelOutputWithPastAndCrossAttentions(
            last_hidden_state=logits,
            past_key_values=past_key_values
        )



import random
from typing import TYPE_CHECKING, Tuple

import torch
from transformers.modeling_outputs import BaseModelOutputWithPastAndCrossAttentions
from transformers.models.t5.modeling_t5 import T5Stack

from mlit.models.common import InputDescription, OnnxModelConverterHelper

if TYPE_CHECKING:
    from onnxruntime import InferenceSession
    import torch.nn


class T5EncoderDescription(OnnxModelConverterHelper[T5Stack]):
    '''
    https://github.com/huggingface/transformers/blob/master/src/transformers/models/t5/modeling_t5.py#L836-L850
    '''
    subname = 'encoder'
    inputs = {
        'input_ids': InputDescription(
            dynamic_axis={0: 'batch_size', 1: 'encoder_seq_len'}
        ),
        'attention_mask': InputDescription(
            dynamic_axis={0: 'batch_size', 1: 'encoder_seq_len'}
        )
    }
    outputs = {
        'last_hidden_state': InputDescription(
            dynamic_axis={0: 'batch_size', 1: 'input_seq_len'}
        )
    }
    default_forward_args = {
        'encoder_hidden_states': None,
        'encoder_attention_mask': None,
        'inputs_embeds': None,
        'head_mask': None,
        # 'cross_attn_layer_head_mask': None,
        'past_key_values': None,
        'use_cache': False,
        'output_attentions': False,
        'output_hidden_states': False,
        'return_dict': True
    }

    def __init__(self, encoder: T5Stack):
        super().__init__(encoder)

    def forward(self, *args, **kwargs):
        result = super().forward(*args, **kwargs)
        return [result.last_hidden_state]

    def sample_inputs(
        self,
        batch_size: int = 2,
        sequence_length: int = 50
    ) -> Tuple:

        vocab_size: int = self.model.config.vocab_size  # type: ignore

        input_ids = torch.randint(low=0,
                                  high=vocab_size - 1,
                                  size=(batch_size, sequence_length),
                                  dtype=torch.int64)

        attention_mask = torch.ones(
            [batch_size, sequence_length],
            dtype=torch.int64)

        if sequence_length >= 2:
            # mask one word in a random position
            padding_position = random.randint(1, sequence_length - 1)
            attention_mask[:, padding_position:] = 0

        return (
            input_ids,
            attention_mask
        )


class T5EncoderInferenceSessionWrapper(T5Stack):
    _onnx_model: 'InferenceSession'

    def __init__(self, onnx_model: 'InferenceSession', config):
        super(T5Stack, self).__init__(config)
        self._onnx_model = onnx_model

    def forward(self,
                input_ids=None,
                attention_mask=None,
                return_dict=None,
                **kwargs):
        [hidden_states] = self._onnx_model.run(
            None,
            {
                'input_ids': input_ids.cpu().numpy(),
                'attention_mask': attention_mask.cpu().numpy()
            }
        )
        hidden_states = torch.FloatTensor(hidden_states)

        if not return_dict:
            return tuple(hidden_states)
        return BaseModelOutputWithPastAndCrossAttentions(
            last_hidden_state=hidden_states,
        )



import os
import os.path
from pathlib import Path

from onnxruntime.quantization import quantize_dynamic, quantize_qat, quantize_static

from mlit.models.common import OnnxModelConverterHelper


def quantize_model(
    model_helper: OnnxModelConverterHelper,
    output_base_dir: str
) -> str:
    output_base_dir = os.path.abspath(output_base_dir)
    assert os.path.isdir(
        output_base_dir), f'No such base directory: {output_base_dir}'

    export_dir = os.path.join(output_base_dir, model_helper.name)
    assert os.path.isdir(
        export_dir), f'No such export directory: {export_dir}'

    input_file = os.path.join(export_dir, f'{model_helper.subname}.onnx')
    assert os.path.isfile(
        input_file), f'No such input file found: {input_file}'

    quantize_dir = os.path.join(export_dir, 'quantized')
    if not os.path.isdir(quantize_dir):
        os.mkdir(quantize_dir)

    output_file = os.path.join(quantize_dir, os.path.basename(input_file))

    method = model_helper.onnx_quantize_method
    args = model_helper.onnx_quantize_args

    input_file = Path(input_file)
    output_file = Path(output_file)

    if method == 'quantize_static':
        out_model = quantize_static(input_file, output_file, ** args)
    elif method == 'quantize_dynamic':
        out_model = quantize_dynamic(input_file, output_file, **args)
    elif method == 'quantize_qat':
        out_model = quantize_qat(input_file, output_file, **args)
    else:
        assert False, f'Unknown quantization method: {method}'

    return output_file.as_posix()



from argparse import ArgumentParser, Namespace
from enum import Enum
import logging
import sys
import onnxruntime

from transformers import PreTrainedTokenizerFast, pipeline


from mlit.models.t5.helpers.loader import load_inference_model, load_model as load_model_t5
from mlit.onnx_support.export import export_config, export_model

logger = logging.getLogger()
logger.setLevel(logging.WARN)

sh = logging.StreamHandler(sys.stdout)
sh.setLevel(logging.WARN)
sh.setFormatter(logging.Formatter(
    '%(asctime)s [%(levelname)s] %(message)s'))
logger.addHandler(sh)


class Task(Enum):
    TO_ONNX = 'to-onnx'
    INFERENCE = 'inference'

    def __str__(self):
        return self.value


class ModelType(Enum):
    T5 = 't5'

    def __str__(self):
        return self.value


def to_onnx(args: Namespace):
    model_type = ModelType(args.model_type)
    helpers = []
    if model_type == ModelType.T5:
        helpers, config = load_model_t5(
            args.model_name_or_path,
            args.config_name,
            args.cache_dir
        )
    else:
        assert False, f'Model type not supported: {model_type}'

    for helper in helpers:
        file_paths = export_model(
            helper,
            args.export_dir,
            args.do_quantize
        )
        logger.info(f'Created files: {file_paths}')
    export_config(
        config, args.export_dir, helpers[0].name
    )


def inference(args: Namespace):
    model_type = ModelType(args.model_type)
    if model_type == ModelType.T5:
        # inference_options = onnxruntime.SessionOptions()
        # inference_options.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_ENABLE_ALL
        # inference_options.enable_profiling = True
        logger.info('Loading inference model')
        model = load_inference_model(
            args.model_name,
            args.base_dir,
            args.model_subtype,
            args.quantized,
            use_threading=True,
            use_compressed_files=args.compressed,
            # options=inference_options
        )
        logger.info('Loaded inference model')

        # TODO: temporary
        tokenizer = PreTrainedTokenizerFast.from_pretrained(
            args.tokenizer_name or args.model_name)
        model.__class__.__name__ = 'T5ForConditionalGeneration'
        pl = pipeline(
            'text2text-generation',
            model=model,
            tokenizer=tokenizer
        )
        logger.info('Started generation pipeline')
        result = pl(args.model_input, use_cache=True, max_length=50)
        logger.info('Finished generation pipeline')
        print(result)


def get_subparser_to_onnx(parser: ArgumentParser) -> ArgumentParser:
    parser.add_argument('--model-type', dest='model_type', required=True, type=str,
                        choices=list(map(lambda x: x.value, ModelType)), help='Model type')
    parser.add_argument('--model-name', dest='model_name_or_path', required=True,
                        type=str, help='Model name or path')
    parser.add_argument('--export-dir', dest='export_dir', type=str, required=True,
                        help='Base export directory')
    parser.add_argument('--cache-dir', dest='cache_dir', type=str, required=False,
                        default=None, help='Transformers cache directory')
    parser.add_argument('--config-name', dest='config_name', type=str, required=False,
                        default=None, help='Optional config name (if different from model)')
    parser.add_argument('--quantize', dest='do_quantize', type=bool, required=False,
                        default=True, help='Quantize or not')

    parser.set_defaults(func=to_onnx)
    return parser


def get_subparser_inference(parser: ArgumentParser) -> ArgumentParser:
    parser.add_argument('--model-name', dest='model_name', required=True,
                        type=str, help='Model name')
    parser.add_argument('--base-dir', dest='base_dir', type=str, required=True,
                        help='Base export directory')
    parser.add_argument('--model-type', dest='model_type', required=True, type=str,
                        choices=list(map(lambda x: x.value, ModelType)), help='Model type')
    parser.add_argument('--model-subtype', dest='model_subtype', type=str, default='history',
                        choices=['history', 'no_history'], help='Model type')
    parser.add_argument('--quantized', dest='quantized', type=bool, required=False,
                        default=True, help='Quantized or not')
    parser.add_argument('--compressed', dest='compressed', required=False,
                        default=False, action='store_true', help='Quantized or not')

    parser.add_argument('--tokenizer-name', dest='tokenizer_name', type=str, default=None,
                        required=False, help='Optional tokenizer name (model name by default)')
    parser.add_argument('--model-input', dest='model_input', type=str, required=True,
                        help='Model input')

    parser.set_defaults(func=inference)

    return parser


def get_parser() -> ArgumentParser:
    parser = ArgumentParser(
        description='ML Inference tools CLI.',
    )

    subparsers = parser.add_subparsers(help='sub-command help')

    subparser = subparsers.add_parser(str(Task.TO_ONNX), help='onnx help')
    get_subparser_to_onnx(subparser)

    subparser = subparsers.add_parser(
        str(Task.INFERENCE), help='inference help')
    get_subparser_inference(subparser)

    parser.add_argument('--verbose', dest='is_verbose', required=False,
                        default=False, action='store_true',
                        help='Enable verbose logging')

    parser.add_argument('--debug', dest='is_debug', required=False,
                        default=False, action='store_true',
                        help='Enable debug logging')

    return parser


if __name__ == '__main__':
    args = get_parser().parse_args()
    if args.is_verbose:
        logger.setLevel(logging.INFO)
        sh.setLevel(logging.INFO)
    if args.is_debug:
        logger.setLevel(logging.DEBUG)
        sh.setLevel(logging.DEBUG)
    args.func(args)



import copy

import torch
from transformers import T5ForConditionalGeneration, T5PreTrainedModel

from .decoder import (
    T5DecoderInferenceSessionWrapper,
    T5DecoderNoHistoryInferenceSessionWrapper,
)
from .encoder import T5EncoderInferenceSessionWrapper


class OnnxT5LMHeadModelNoHistory(T5ForConditionalGeneration):
    def __init__(self, config, encoder_session, decoder_session):
        T5PreTrainedModel.__init__(self, config)  # type: ignore

        self.model_dim = config.d_model

        encoder_config = copy.deepcopy(config)
        encoder_config.is_decoder = False
        encoder_config.use_cache = False
        encoder_config.is_encoder_decoder = False
        self.encoder = T5EncoderInferenceSessionWrapper(
            encoder_session, encoder_config)

        decoder_config = copy.deepcopy(config)
        decoder_config.is_decoder = True
        decoder_config.is_encoder_decoder = False
        decoder_config.num_layers = config.num_decoder_layers
        self.decoder = T5DecoderNoHistoryInferenceSessionWrapper(
            decoder_session, decoder_config)

        # LM head is already integrated into decoder.
        # set it to pass through and disable all logit math in the original model
        self.lm_head = lambda x: x
        self.config.tie_word_embeddings = False  # type: ignore

        # Model parallel
        self.model_parallel = False
        self.device_map = None

    @property
    def device(self):
        return torch.device('cpu')


class OnnxT5LMHeadModel(T5ForConditionalGeneration):
    def __init__(self, config, encoder_session, decoder_session_first_step, decoder_session):
        T5PreTrainedModel.__init__(self, config)  # type: ignore

        self.model_dim = config.d_model

        encoder_config = copy.deepcopy(config)
        encoder_config.is_decoder = False
        encoder_config.use_cache = False
        encoder_config.is_encoder_decoder = False
        self.encoder = T5EncoderInferenceSessionWrapper(
            encoder_session, encoder_config)

        decoder_config = copy.deepcopy(config)
        decoder_config.is_decoder = True
        decoder_config.is_encoder_decoder = False
        decoder_config.num_layers = config.num_decoder_layers
        self.decoder = T5DecoderInferenceSessionWrapper(
            decoder_session_first_step, decoder_session, decoder_config)

        # LM head is already integrated into decoder.
        # set it to pass through and disable all logit math in the original model
        self.lm_head = lambda x: x
        self.config.tie_word_embeddings = False  # type: ignore

        # Model parallel
        self.model_parallel = False
        self.device_map = None

    @property
    def device(self):
        return torch.device('cpu')



from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Dict, Generic, Optional, Tuple, TypeVar

import torch.nn
from typing_extensions import Literal

if TYPE_CHECKING:
    from transformers.configuration_utils import PretrainedConfig

T = TypeVar('T', bound=torch.nn.Module)


@dataclass
class InputDescription:
    dynamic_axis: Optional[Dict[int, str]]


class OnnxModelConverterHelper(torch.nn.Module, Generic[T]):
    _model: T
    inputs: Dict[str, InputDescription]
    outputs: Dict[str, InputDescription]
    default_forward_args: Dict[str, Any] = {}
    subname: str = 'model'

    def __init__(self, model: T):
        super().__init__()
        self._model = model

    @property
    def model(self):
        return self._model

    @property
    def config(self) -> 'PretrainedConfig':
        return self.model.config  # type: ignore

    @property
    def name(self) -> str:
        name: str = self.model.config.name_or_path  # type: ignore
        name = name.split('/')[-1]
        return name

    def forward(self, *args, **kwargs):
        return self.model.forward(
            *args,
            **{**self.default_forward_args, **kwargs}
        )

    def sample_inputs(self) -> Tuple:
        return ()

    def get_onnx_export_args(self) -> Dict[str, Any]:
        dynamic_axes = {}
        for name, desc in self.inputs.items():
            if desc.dynamic_axis is not None:
                dynamic_axes[name] = desc.dynamic_axis
        for name, desc in self.outputs.items():
            if desc.dynamic_axis is not None:
                dynamic_axes[name] = desc.dynamic_axis

        return {
            'model': self,
            'args': self.sample_inputs(),
            'input_names': list(self.inputs.keys()),
            'output_names': list(self.outputs.keys()),
            'dynamic_axes': dynamic_axes,
            'opset_version': 12,
            'export_params': True,
            'do_constant_folding': True
        }

    @property
    def onnx_quantize_method(self) -> Literal['quantize_static', 'quantize_dynamic', 'quantize_qat']:
        return 'quantize_qat'

    @property
    def onnx_quantize_args(self) -> Dict[str, Any]:
        from onnxruntime.quantization import QuantType
        return {
            'weight_type': QuantType.QInt8,
            'op_types_to_quantize': ['MatMul']
        }



import logging
import os
import os.path
from typing import TYPE_CHECKING, Tuple, Union

import torch.onnx

from mlit.models.common import OnnxModelConverterHelper

if TYPE_CHECKING:
    from transformers import PretrainedConfig


logger = logging.getLogger(__name__)


def export_config(
    config: 'PretrainedConfig',
    output_base_dir: str,
    model_name
):
    config_dir = os.path.join(
        output_base_dir, model_name, 'transformers_config')
    if not os.path.isdir(config_dir):
        os.mkdir(config_dir)

    config.save_pretrained(config_dir)


def export_model(
    model_helper: OnnxModelConverterHelper,
    output_base_dir: str,
    do_quantize: bool = False
) -> Union[str, Tuple[str, str]]:
    output_base_dir = os.path.abspath(output_base_dir)
    assert os.path.isdir(
        output_base_dir), f'No such base directory: {output_base_dir}'

    export_dir = os.path.join(output_base_dir, model_helper.name)
    if not os.path.isdir(export_dir):
        os.mkdir(export_dir)

    export_file = os.path.join(export_dir, f'{model_helper.subname}.onnx')

    args = model_helper.get_onnx_export_args()
    args['f'] = export_file
    logger.debug(f'Exporting {export_file} with {args["input_names"]}')

    torch.onnx.export(**args)

    if do_quantize:
        from mlit.onnx_support.quantize import quantize_model

        quantized_file = quantize_model(model_helper, output_base_dir)
        return (export_file, quantized_file)
    else:
        return export_file




import logging
import io
from typing import List, Optional, Union
import onnxruntime
from concurrent.futures import ThreadPoolExecutor, as_completed


logger = logging.getLogger(__name__)


def load_inference_session(
    file_paths: Union[str, List[str]],
    options: Optional[onnxruntime.SessionOptions] = None,
    use_threading=True
):

    def _create_session(file_path: str) -> onnxruntime.InferenceSession:
        if file_path.endswith('.zst'):
            from pyzstd import decompress_stream

            logger.debug(f'Reading compressed model file: {file_path}')
            with io.open(file_path, 'rb') as ifh:
                with io.BytesIO() as bo:
                    decompress_stream(ifh, bo)
                    model_bytes = bo.getvalue()
            logger.debug(
                f'Creating session from compressed model file: {file_path}')
            sess = onnxruntime.InferenceSession(model_bytes, options)
            # Free memory
            # https://github.com/microsoft/onnxruntime/blob/master/onnxruntime/python/onnxruntime_inference_collection.py#L270
            sess._model_bytes = None
            logger.debug(
                f'Session created from compressed model file: {file_path}')
            return sess
        else:
            logger.debug(f'Creating session from model file: {file_path}')
            sess = onnxruntime.InferenceSession(file_path, options)
            logger.debug(f'Session created from model file: {file_path}')
            return sess

    session_file_paths = [file_paths] if isinstance(
        file_paths, str) else file_paths

    if use_threading and len(session_file_paths) > 1:
        with ThreadPoolExecutor() as executor:
            futures = []
            for idx, path in enumerate(session_file_paths):
                futures.append(executor.submit(
                    lambda p: (idx, _create_session(p)), p=path))
            sessions = [future.result() for future in as_completed(futures)]
            sessions = sorted(sessions, key=lambda s: s[0])
            sessions = [s[1] for s in sessions]
    else:
        sessions = [_create_session(p) for p in session_file_paths]

    if isinstance(file_paths, str):
        return sessions[0]
    else:
        return sessions



import os.path
import sys

import setuptools

from mlit import __description__, __version__

PROJECT_DIR = os.path.abspath(os.path.dirname(__file__))

if PROJECT_DIR not in sys.path:
    sys.path.append(PROJECT_DIR)


with open('README.md', 'r') as fh:
    long_description = fh.read()


setup_args = dict(
    name='mlit',
    version=__version__,
    # url=None,
    author='roman@kalyakin.com',
    description=__description__,
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=[
        'transformers[torch]>=4.4.0.dev0',
        'onnxruntime>=1.7.0'
    ],
    zip_safe=False,
    include_package_data=True,
    python_requires=">=3.7",
    # license=None,
    platforms="Linux, Mac OS X, Windows",
    keywords=["ML", "hugginface", "onnx"],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7"
    ],
)

if __name__ == "__main__":
    setuptools.setup(**setup_args)




import io
import logging
import os
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import TYPE_CHECKING, List, Optional, Tuple, Union, cast

from mlit.models.t5.helpers.decoder import (
    T5DecoderFirstStepHistoryDescription, T5DecoderHistoryDescription,
    T5DecoderNoHistoryDescription)
from mlit.models.t5.helpers.encoder import T5EncoderDescription
from mlit.models.t5.helpers.model import (OnnxT5LMHeadModel,
                                          OnnxT5LMHeadModelNoHistory)
from onnxruntime import InferenceSession, SessionOptions
from transformers import T5Config, T5ForConditionalGeneration
from typing_extensions import Literal

if TYPE_CHECKING:
    from mlit.models.common import OnnxModelConverterHelper
    from torch.nn.modules.linear import Linear
    from transformers.models.t5.modeling_t5 import T5Stack

logger = logging.getLogger(__name__)


def load_model(
    name_or_path: str,
    config_name: str = None,
    cache_dir: str = None
) -> Tuple[List['OnnxModelConverterHelper'], T5Config]:

    model_config: Optional[T5Config] = None
    if config_name is not None:
        model_config = cast(T5Config, T5Config.from_pretrained(config_name))

    model: T5ForConditionalGeneration = T5ForConditionalGeneration.from_pretrained(
        name_or_path, config=model_config, cache_dir=cache_dir).eval()  # type: ignore

    encoder: T5Stack = model.encoder  # type: ignore
    decoder: T5Stack = model.decoder  # type: ignore
    lm_head: Linear = model.lm_head  # type: ignore
    config: T5Config = model.config  # type: ignore

    return [
        T5EncoderDescription(encoder),
        T5DecoderNoHistoryDescription(decoder, lm_head),
        T5DecoderFirstStepHistoryDescription(decoder, lm_head),
        T5DecoderHistoryDescription(decoder, lm_head)
    ], config


def _as_quantized(path_elements: List[str]) -> List[str]:
    return path_elements[:-1] + ['quantized'] + path_elements[-1:]


def load_inference_model(
    name: str,
    base_dir: str,
    model_type: Literal['no_history', 'history'],
    quantized: bool,
    use_threading: bool = True,
    use_compressed_files: bool = False,
    options: Optional[SessionOptions] = None
) -> T5ForConditionalGeneration:
    config_location = os.path.join(
        base_dir, name, 'transformers_config')
    config = T5Config.from_pretrained(config_location)

    base_path = [base_dir, name]

    file_suffix = 'onnx' if not use_compressed_files else 'onnx.zst'

    def construct_path(filename: str) -> str:
        parts = base_path + [filename]
        if quantized:
            parts = _as_quantized(parts)
        return os.path.join(*parts)

    if model_type == 'no_history':
        encoder_path = construct_path(
            f'{T5EncoderDescription.subname}.{file_suffix}')
        decoder_path = construct_path(
            f'{T5DecoderNoHistoryDescription.subname}.{file_suffix}')

        klass = OnnxT5LMHeadModelNoHistory
        klass_config = config
        klass_session_paths = [encoder_path, decoder_path]

    elif model_type == 'history':
        encoder_path = construct_path(
            f'{T5EncoderDescription.subname}.{file_suffix}')
        decoder_first_step_path = construct_path(
            f'{T5DecoderFirstStepHistoryDescription.subname}.{file_suffix}')
        decoder_path = construct_path(
            f'{T5DecoderHistoryDescription.subname}.{file_suffix}')

        klass = OnnxT5LMHeadModel
        klass_config = config
        klass_session_paths = [encoder_path,
                               decoder_first_step_path, decoder_path]

    else:
        raise Exception(f'Unknown model type: {model_type}')

    def _create_session(file_path: str, options: SessionOptions) -> InferenceSession:
        if use_compressed_files:
            # import lzma
            from pyzstd import decompress_stream

            logger.debug(f'Reading compressed model file: {file_path}')
            with io.open(file_path, 'rb') as ifh:
                with io.BytesIO() as bo:
                    decompress_stream(ifh, bo)
                    model_bytes = bo.getvalue()
            # with lzma.open(file_path, 'rb') as f:
            #     model_bytes = f.read()
            logger.debug(
                f'Creating session from compressed model file: {file_path}')
            sess = InferenceSession(model_bytes, options)
            # Free memory
            # https://github.com/microsoft/onnxruntime/blob/master/onnxruntime/python/onnxruntime_inference_collection.py#L270
            sess._model_bytes = None
            logger.debug(
                f'Session created from compressed model file: {file_path}')
            return sess
        else:
            logger.debug(f'Creating session from model file: {file_path}')
            sess = InferenceSession(file_path, options)
            logger.debug(f'Session created from model file: {file_path}')
            return sess

    if use_threading:
        with ThreadPoolExecutor() as executor:
            futures = []
            for idx, path in enumerate(klass_session_paths):
                futures.append(executor.submit(
                    lambda p: (idx, _create_session(p, options)), p=path))
            sessions = [future.result() for future in as_completed(futures)]
            sessions = sorted(sessions, key=lambda s: s[0])
            sessions = [s[1] for s in sessions]
    else:
        sessions = [_create_session(p, options) for p in klass_session_paths]

    return klass(klass_config, *sessions)


def load_inference_model_from_files(
    filenames: Union[Tuple[str, str], Tuple[str, str, str]],
    transformers_config_path: str,
    options: Optional[SessionOptions] = None
) -> T5ForConditionalGeneration:
    config = T5Config.from_pretrained(transformers_config_path)

    if len(filenames) == 2:
        klass = OnnxT5LMHeadModelNoHistory
    elif len(filenames) == 3:
        klass = OnnxT5LMHeadModel
    else:
        raise Exception(f'Unexpected number of filenames: {len(filenames)}')

    def _create_session(file_path: str, options: SessionOptions) -> InferenceSession:
        if file_path.endswith('.zst'):
            from pyzstd import decompress_stream

            logger.debug(f'Reading compressed model file: {file_path}')
            with io.open(file_path, 'rb') as ifh:
                with io.BytesIO() as bo:
                    decompress_stream(ifh, bo)
                    model_bytes = bo.getvalue()
            logger.debug(
                f'Creating session from compressed model file: {file_path}')
            sess = InferenceSession(model_bytes, options)
            # Free memory
            # https://github.com/microsoft/onnxruntime/blob/master/onnxruntime/python/onnxruntime_inference_collection.py#L270
            sess._model_bytes = None
            logger.debug(
                f'Session created from compressed model file: {file_path}')
            return sess
        else:
            logger.debug(f'Creating session from model file: {file_path}')
            sess = InferenceSession(file_path, options)
            logger.debug(f'Session created from model file: {file_path}')
            return sess

    with ThreadPoolExecutor() as executor:
        futures = []
        for idx, path in enumerate(filenames):
            futures.append(executor.submit(
                lambda p: (idx, _create_session(p, options)), p=path))
        sessions = [future.result() for future in as_completed(futures)]
        sessions = sorted(sessions, key=lambda s: s[0])
        sessions = [s[1] for s in sessions]

    return klass(config, *sessions)
